from pathlib import Path
from pyfpga.factory import Factory

tdir = Path(__file__).parent.resolve()


def test_diamond():
    tool = 'diamond'
    generate(tool, 'PARTNAME')
    base = f'results/{tool}/{tool}'
    assert Path(f'{base}.tcl').exists(), 'file not found'
    assert Path(f'{base}-prog.sh').exists(), 'file not found'


def test_ise():
    tool = 'ise'
    generate(tool, 'DEVICE-PACKAGE-SPEED')
    base = f'results/{tool}/{tool}'
    assert Path(f'{base}.tcl').exists(), 'file not found'
    assert Path(f'{base}-prog.tcl').exists(), 'file not found'


def test_libero():
    tool = 'libero'
    generate(tool, 'DEVICE-PACKAGE-SPEED')
    base = f'results/{tool}/{tool}'
    assert Path(f'{base}.tcl').exists(), 'file not found'
    assert Path(f'{base}-prog.tcl').exists(), 'file not found'


def test_openflow():
    tool = 'openflow'
    generate(tool, 'DEVICE-PACKAGE')
    base = f'results/{tool}/{tool}'
    assert Path(f'{base}.sh').exists(), 'file not found'
    assert Path(f'{base}-prog.sh').exists(), 'file not found'


def test_quartus():
    tool = 'quartus'
    generate(tool, 'PARTNAME')
    base = f'results/{tool}/{tool}'
    assert Path(f'{base}.tcl').exists(), 'file not found'
    assert Path(f'{base}-prog.tcl').exists(), 'file not found'


def test_vivado():
    tool = 'vivado'
    generate(tool, 'PARTNAME')
    base = f'results/{tool}/{tool}'
    assert Path(f'{base}.tcl').exists(), 'file not found'
    assert Path(f'{base}-prog.tcl').exists(), 'file not found'


def generate(tool, part):
    prj = Factory(tool, odir=f'results/{tool}')
    prj.set_part(part)
    prj.set_top('TOPNAME')
    prj.add_include(str(tdir / 'fakedata/dir1'))
    prj.add_include(str(tdir / 'fakedata/dir2'))
    if tool != 'ise':
        prj.add_slog(str(tdir / 'fakedata/**/*.sv'))
    prj.add_vhdl(str(tdir / 'fakedata/**/*.vhdl'), 'LIB')
    prj.add_vlog(str(tdir / 'fakedata/**/*.v'))
    prj.add_cons(str(tdir / 'fakedata/cons/all.xdc'))
    prj.add_cons(str(tdir / 'fakedata/cons/syn.xdc'))
    prj.add_cons(str(tdir / 'fakedata/cons/par.xdc'))
    prj.add_param('PAR1', 'VAL1')
    prj.add_param('PAR2', 'VAL2')
    prj.add_define('DEF1', 'VAL1')
    prj.add_define('DEF2', 'VAL2')
    prj.add_hook('precfg', 'HOOK01')
    prj.add_hook('precfg', 'HOOK02')
    prj.add_hook('postcfg', 'HOOK03')
    prj.add_hook('postcfg', 'HOOK04')
    prj.add_hook('presyn', 'HOOK05')
    prj.add_hook('presyn', 'HOOK06')
    prj.add_hook('postsyn', 'HOOK07')
    prj.add_hook('postsyn', 'HOOK08')
    prj.add_hook('prepar', 'HOOK09')
    prj.add_hook('prepar', 'HOOK10')
    prj.add_hook('postpar', 'HOOK11')
    prj.add_hook('postpar', 'HOOK12')
    prj.add_hook('prebit', 'HOOK13')
    prj.add_hook('prebit', 'HOOK14')
    prj.add_hook('postbit', 'HOOK15')
    prj.add_hook('postbit', 'HOOK16')
    try:
        prj.make()
    except RuntimeError:
        pass
    if tool == 'libero':
        open(f'results/{tool}/{tool}.ppd', 'w').close()
    elif tool == 'quartus':
        open(f'results/{tool}/{tool}.sof', 'w').close()
    else:
        open(f'results/{tool}/{tool}.bit', 'w').close()
    try:
        prj.prog()
    except RuntimeError:
        pass



#
# Copyright (C) 2020-2024 PyFPGA Project
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""
Implements support for an Open Source development flow.
"""

from pyfpga.project import Project


class Openflow(Project):
    """Class to support Open Source tools."""

    def _configure(self):
        tool = 'openflow'
        self.conf['tool'] = tool
        self.conf['make_cmd'] = f'bash {tool}.sh'
        self.conf['make_ext'] = 'sh'
        self.conf['prog_bit'] = ['svf', 'bit']
        self.conf['prog_cmd'] = f'bash {tool}-prog.sh'
        self.conf['prog_ext'] = 'sh'

    def _make_custom(self):
        info = get_info(self.data.get('part', 'hx8k-ct256'))
        self.data['family'] = info['family']
        self.data['device'] = info['device']
        self.data['package'] = info['package']

    def _prog_custom(self):
        info = get_info(self.data.get('part', 'hx8k-ct256'))
        self.data['family'] = info['family']


def get_info(part):
    """Get info about the FPGA part.

    :param part: the FPGA part as specified by the tool
    :returns: a dict with the keys family, device and package
    """
    part = part.lower().replace(' ', '')
    # Looking for the family
    family = None
    families = [
        # From <YOSYS>/techlibs/xilinx/synth_xilinx.cc
        'xcup', 'xcu', 'xc7', 'xc6s', 'xc6v', 'xc5v', 'xc4v', 'xc3sda',
        'xc3sa', 'xc3se', 'xc3s', 'xc2vp', 'xc2v', 'xcve', 'xcv'
    ]
    for item in families:
        if part.startswith(item):
            family = item
            break
    families = [
        # From <nextpnr>/ice40/main.cc
        'lp384', 'lp1k', 'lp4k', 'lp8k', 'hx1k', 'hx4k', 'hx8k',
        'up3k', 'up5k', 'u1k', 'u2k', 'u4k'
    ]
    if part.startswith(tuple(families)):
        family = 'ice40'
    families = [
        # From <nextpnr>/ecp5/main.cc
        '12k', '25k', '45k', '85k', 'um-25k', 'um-45k', 'um-85k',
        'um5g-25k', 'um5g-45k', 'um5g-85k'
    ]
    if part.startswith(tuple(families)):
        family = 'ecp5'
    # Looking for the other values
    device = None
    package = None
    aux = part.split('-')
    if len(aux) == 2:
        device = aux[0]
        package = aux[1]
    elif len(aux) == 3:
        device = f'{aux[0]}-{aux[1]}'
        package = aux[2]
    else:
        valid = 'DEVICE-PACKAGE'
        raise ValueError(f'Invalid PART format ({valid})')
    if family in ['lp4k', 'hx4k']:  # See http://www.clifford.at/icestorm
        device = device.replace('4', '8')
        package += ":4k"
    if family == 'ecp5':
        package = package.upper()
    # Finish
    return {
        'family': family,
        'device': device,
        'package': package
    }



#
# Copyright (C) 2019-2024 PyFPGA Project
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""
Implements support for Libero.
"""

import re

from pyfpga.project import Project


class Libero(Project):
    """Class to support Libero projects."""

    def _configure(self):
        tool = 'libero'
        self.conf['tool'] = tool
        self.conf['make_cmd'] = f'{tool} SCRIPT:{tool}.tcl'
        self.conf['make_ext'] = 'tcl'
        self.conf['prog_bit'] = ['ppd', 'stp', 'bit', 'jed']
        self.conf['prog_cmd'] = f'FPExpress SCRIPT:{tool}-prog.tcl'
        self.conf['prog_ext'] = 'tcl'

    def _make_custom(self):
        info = get_info(self.data.get('part', 'mpf100t-1-fcg484'))
        self.data['family'] = info['family']
        self.data['device'] = info['device']
        self.data['speed'] = info['speed']
        self.data['package'] = info['package']
        self.data['prange'] = info['prange']


# pylint: disable=duplicate-code

def get_info(part):
    """Get info about the FPGA part.

    :param part: the FPGA part as specified by the tool
    :returns: a dict with the keys family, device, speed, package and prange
    """
    part = part.lower().replace(' ', '')
    # Looking for the family
    family = None
    families = {
        r'm2s': 'SmartFusion2',
        r'm2gl': 'IGLOO2',
        r'rt4g': 'RTG4',
        r'mpfs': 'PolarFireSoC',
        r'mpf': 'PolarFire',
        r'a2f': 'SmartFusion',
        r'afs': 'Fusion',
        r'aglp': 'IGLOO+',
        r'agle': 'IGLOOE',
        r'agl': 'IGLOO',
        r'a3p\d+l': 'ProAsic3L',
        r'a3pe': 'ProAsic3E',
        r'a3p': 'ProAsic3'
    }
    for key, value in families.items():
        if re.match(key, part):
            family = value
            break
    # Looking for the other values
    device = None
    speed = None
    package = None
    prange = None
    aux = part.split('-')
    if len(aux) == 2:
        device = aux[0]
        package = aux[1]
        if package[0].isdigit():
            speed = f'-{package[0]}'
            package = package[1:]
        else:
            speed = 'STD'
    elif len(aux) == 3:
        device = aux[0]
        if len(aux[1]) < len(aux[2]):
            speed = f'-{aux[1]}'
            package = aux[2]
        else:
            speed = f'-{aux[2]}'
            package = aux[1]
    else:
        valid = 'DEVICE-[SPEED][-]PACKAGE[PRANGE][-SPEED]'
        raise ValueError(f'Invalid PART format ({valid})')
    pranges = {
        'c': 'COM',
        'e': 'EXT',
        'i': 'IND',
        'm': 'MIL',
        't1': 'TGrade1',
        't2': 'TGrade2'
    }
    prange = 'COM'
    for suffix, name in pranges.items():
        if package.endswith(suffix):
            package = package[:-len(suffix)]
            prange = name
            break
    # Finish
    return {
        'family': family,
        'device': device,
        'speed': speed,
        'package': package,
        'prange': prange
    }



#
# Copyright (C) 2019-2024 PyFPGA Project
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""
Implements support for ISE.
"""

import re

from pyfpga.project import Project


class Ise(Project):
    """Class to support ISE projects."""

    def _configure(self):
        tool = 'ise'
        self.conf['tool'] = tool
        self.conf['make_cmd'] = f'xtclsh {tool}.tcl'
        self.conf['make_ext'] = 'tcl'
        self.conf['prog_bit'] = ['bit']
        self.conf['prog_cmd'] = f'impact -batch {tool}-prog.tcl'
        self.conf['prog_ext'] = 'tcl'

    def _make_custom(self):
        info = get_info(self.data.get('part', 'xc7k160t-3-fbg484'))
        self.data['family'] = info['family']
        self.data['device'] = info['device']
        self.data['speed'] = info['speed']
        self.data['package'] = info['package']

    def add_slog(self, pathname):
        """Add System Verilog file/s."""
        raise NotImplementedError('ISE does not support SystemVerilog')


# pylint: disable=duplicate-code

def get_info(part):
    """Get info about the FPGA part.

    :param part: the FPGA part as specified by the tool
    :returns: a dict with the keys family, device, speed and package
    """
    part = part.lower().replace(' ', '')
    # Looking for the family
    family = None
    families = {
        r'xc7a\d+l': 'artix7l',
        r'xc7a': 'artix7',
        r'xc7k\d+l': 'kintex7l',
        r'xc7k': 'kintex7',
        r'xc3sd\d+a': 'spartan3adsp',
        r'xc3s\d+a': 'spartan3a',
        r'xc3s\d+e': 'spartan3e',
        r'xc3s': 'spartan3',
        r'xc6s\d+l': 'spartan6l',
        r'xc6s': 'spartan6',
        r'xc4v': 'virtex4',
        r'xc5v': 'virtex5',
        r'xc6v\d+l': 'virtex6l',
        r'xc6v': 'virtex6',
        r'xc7v\d+l': 'virtex7l',
        r'xc7v': 'virtex7',
        r'xc7z': 'zynq'
    }
    for key, value in families.items():
        if re.match(key, part):
            family = value
            break
    # Looking for the other values
    device = None
    speed = None
    package = None
    aux = part.split('-')
    if len(aux) == 3:
        device = aux[0]
        if len(aux[1]) < len(aux[2]):
            speed = f'-{aux[1]}'
            package = aux[2]
        else:
            speed = f'-{aux[2]}'
            package = aux[1]
    else:
        valid = 'DEVICE-SPEED-PACKAGE or DEVICE-PACKAGE-SPEED'
        raise ValueError(f'Invalid PART format ({valid})')
    # Finish
    return {
        'family': family,
        'device': device,
        'speed': speed,
        'package': package
    }



#
# Copyright (C) 2024 PyFPGA Project
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""
A factory class to create FPGA projects.
"""

# pylint: disable=too-few-public-methods

from pyfpga.diamond import Diamond
from pyfpga.ise import Ise
from pyfpga.libero import Libero
from pyfpga.openflow import Openflow
from pyfpga.quartus import Quartus
from pyfpga.vivado import Vivado


TOOLS = {
    'diamond': Diamond,
    'ise': Ise,
    'libero': Libero,
    'openflow': Openflow,
    'quartus': Quartus,
    'vivado': Vivado
}


class Factory:
    """A factory class to create FPGA projects."""

    def __init__(self, tool='vivado', project=None, odir='results'):
        """Class constructor."""
        if tool not in TOOLS:
            raise NotImplementedError(f'{tool} is unsupported')
        self._instance = TOOLS[tool](project, odir)

    def __getattr__(self, name):
        """Delegate attribute access to the tool instance."""
        return getattr(self._instance, name)



from pathlib import Path

from pyfpga.vivado import Vivado

tdir = Path(__file__).parent.resolve()

pattern = {
    'project': 'EXAMPLE',
    'part': 'PARTNAME',
    'includes': [
        Path(tdir / 'fakedata/dir1').resolve().as_posix(),
        Path(tdir / 'fakedata/dir2').resolve().as_posix(),
        Path(tdir / 'fakedata/dir3').resolve().as_posix()
    ],
    'files': {
        Path(tdir / 'fakedata/vhdl0.vhdl').resolve().as_posix(): {
            'hdl': 'vhdl', 'lib': 'LIB'
        },
        Path(tdir / 'fakedata/dir1/vhdl1.vhdl').resolve().as_posix(): {
            'hdl': 'vhdl', 'lib': 'LIB'
        },
        Path(tdir / 'fakedata/dir2/vhdl2.vhdl').resolve().as_posix(): {
            'hdl': 'vhdl', 'lib': 'LIB'
        },
        Path(tdir / 'fakedata/dir3/vhdl3.vhdl').resolve().as_posix(): {
            'hdl': 'vhdl', 'lib': 'LIB'
        },
        Path(tdir / 'fakedata/vlog0.v').resolve().as_posix(): {
            'hdl': 'vlog'
        },
        Path(tdir / 'fakedata/dir1/vlog1.v').resolve().as_posix(): {
            'hdl': 'vlog'
        },
        Path(tdir / 'fakedata/dir2/vlog2.v').resolve().as_posix(): {
            'hdl': 'vlog'
        },
        Path(tdir / 'fakedata/dir3/vlog3.v').resolve().as_posix(): {
            'hdl': 'vlog'
        },
        Path(tdir / 'fakedata/slog0.sv').resolve().as_posix(): {
            'hdl': 'slog'
        },
        Path(tdir / 'fakedata/dir1/slog1.sv').resolve().as_posix(): {
            'hdl': 'slog'
        },
        Path(tdir / 'fakedata/dir2/slog2.sv').resolve().as_posix(): {
            'hdl': 'slog'
        },
        Path(tdir / 'fakedata/dir3/slog3.sv').resolve().as_posix(): {
            'hdl': 'slog'
        }
    },
    'top': 'TOPNAME',
    'constraints': {
        Path(tdir / 'fakedata/cons/all.xdc').resolve().as_posix(): {},
        Path(tdir / 'fakedata/cons/syn.xdc').resolve().as_posix(): {},
        Path(tdir / 'fakedata/cons/par.xdc').resolve().as_posix(): {}
    },
    'params': {
        'PAR1': 'VAL01',
        'PAR2': 'VAL02',
        'PAR3': 'VAL03'
    },
    'defines': {
        'DEF1': 'VAL01',
        'DEF2': 'VAL02',
        'DEF3': 'VAL03'
    },
    'hooks': {
        'precfg': ['CMD01', 'CMD02'],
        'postcfg': ['CMD03', 'CMD04'],
        'presyn': ['CMD05', 'CMD06'],
        'postsyn': ['CMD07', 'CMD08'],
        'prepar': ['CMD09', 'CMD10'],
        'postpar': ['CMD11', 'CMD12'],
        'prebit': ['CMD13', 'CMD14'],
        'postbit': ['CMD15', 'CMD16']
    }
}


def test_data():
    prj = Vivado('EXAMPLE')
    prj.set_part('PARTNAME')
    prj.set_top('TOPNAME')
    prj.add_include(str(tdir / 'fakedata/dir1'))
    prj.add_include(str(tdir / 'fakedata/dir2'))
    prj.add_include(str(tdir / 'fakedata/dir3'))
    prj.add_slog(str(tdir / 'fakedata/**/*.sv'))
    prj.add_vhdl(str(tdir / 'fakedata/**/*.vhdl'), 'LIB')
    prj.add_vlog(str(tdir / 'fakedata/**/*.v'))
    prj.add_cons(str(tdir / 'fakedata/cons/all.xdc'))
    prj.add_cons(str(tdir / 'fakedata/cons/syn.xdc'))
    prj.add_cons(str(tdir / 'fakedata/cons/par.xdc'))
    prj.add_param('PAR1', 'VAL01')
    prj.add_param('PAR2', 'VAL02')
    prj.add_param('PAR3', 'VAL03')
    prj.add_define('DEF1', 'VAL01')
    prj.add_define('DEF2', 'VAL02')
    prj.add_define('DEF3', 'VAL03')
    prj.add_hook('precfg', 'CMD01')
    prj.add_hook('precfg', 'CMD02')
    prj.add_hook('postcfg', 'CMD03')
    prj.add_hook('postcfg', 'CMD04')
    prj.add_hook('presyn', 'CMD05')
    prj.add_hook('presyn', 'CMD06')
    prj.add_hook('postsyn', 'CMD07')
    prj.add_hook('postsyn', 'CMD08')
    prj.add_hook('prepar', 'CMD09')
    prj.add_hook('prepar', 'CMD10')
    prj.add_hook('postpar', 'CMD11')
    prj.add_hook('postpar', 'CMD12')
    prj.add_hook('prebit', 'CMD13')
    prj.add_hook('prebit', 'CMD14')
    prj.add_hook('postbit', 'CMD15')
    prj.add_hook('postbit', 'CMD16')
    assert prj.data == pattern, 'ERROR: unexpected data'
    paths = prj.data['includes'] + list(prj.data['files'].keys())
    for path in paths:
        assert '\\' not in path, (
            f"'{path}' contains a '\\' character, which is not allowed."
        )



from pyfpga.ise import get_info as get_info_ise
from pyfpga.libero import get_info as get_info_libero
from pyfpga.openflow import get_info as get_info_openflow


def test_ise():
    info = {
        'family': 'kintex7',
        'device': 'xc7k160t',
        'speed': '-3',
        'package': 'fbg484'
    }
    assert get_info_ise('xc7k160t-3-fbg484') == info
    assert get_info_ise('xc7k160t-fbg484-3') == info
    info['speed'] = '-3l'
    assert get_info_ise('xc7k160t-3L-fbg484') == info
    assert get_info_ise('xc7k160t-fbg484-3L') == info


def test_libero():
    info = {
        'family': 'SmartFusion2',
        'device': 'm2s025t',
        'speed': 'STD',
        'package': 'fg484',
        'prange': 'COM'
    }
    assert get_info_libero('M2S025T-FG484') == info
    info['prange'] = 'IND'
    assert get_info_libero('M2S025T-FG484I') == info
    info['speed'] = '-1'
    info['prange'] = 'COM'
    assert get_info_libero('M2S025T-1FG484') == info
    assert get_info_libero('M2S025T-1-FG484') == info
    assert get_info_libero('M2S025T-FG484-1') == info
    info['prange'] = 'IND'
    assert get_info_libero('M2S025T-1FG484I') == info
    assert get_info_libero('M2S025T-1-FG484I') == info
    assert get_info_libero('M2S025T-FG484I-1') == info
    info['prange'] = 'MIL'
    assert get_info_libero('M2S025T-1FG484M') == info
    assert get_info_libero('M2S025T-1-FG484M') == info
    assert get_info_libero('M2S025T-FG484M-1') == info
    info = {
        'family': 'IGLOO2',
        'device': 'm2gl025',
        'speed': 'STD',
        'package': 'fg484',
        'prange': 'COM'
    }
    assert get_info_libero('M2GL025-FG484') == info
    info['prange'] = 'IND'
    assert get_info_libero('M2GL025-FG484I') == info
    info['speed'] = '-1'
    info['prange'] = 'COM'
    assert get_info_libero('M2GL025-1FG484') == info
    assert get_info_libero('M2GL025-1-FG484') == info
    assert get_info_libero('M2GL025-FG484-1') == info
    info['prange'] = 'IND'
    assert get_info_libero('M2GL025-1FG484I') == info
    assert get_info_libero('M2GL025-1-FG484I') == info
    assert get_info_libero('M2GL025-FG484I-1') == info
    info['prange'] = 'MIL'
    assert get_info_libero('M2GL025-1FG484M') == info
    assert get_info_libero('M2GL025-1-FG484M') == info
    assert get_info_libero('M2GL025-FG484M-1') == info
    info['prange'] = 'TGrade1'
    assert get_info_libero('M2GL025-1FG484T1') == info
    assert get_info_libero('M2GL025-1-FG484T1') == info
    assert get_info_libero('M2GL025-FG484T1-1') == info
    info = {
        'family': 'PolarFire',
        'device': 'mpf300ts_es',
        'speed': 'STD',
        'package': 'fg484',
        'prange': 'EXT'
    }
    assert get_info_libero('MPF300TS_ES-FG484E') == info
    info['prange'] = 'IND'
    assert get_info_libero('MPF300TS_ES-FG484I') == info
    info['speed'] = '-1'
    info['prange'] = 'EXT'
    assert get_info_libero('MPF300TS_ES-1FG484E') == info
    assert get_info_libero('MPF300TS_ES-1-FG484E') == info
    assert get_info_libero('MPF300TS_ES-FG484E-1') == info
    info['prange'] = 'IND'
    assert get_info_libero('MPF300TS_ES-1FG484I') == info
    assert get_info_libero('MPF300TS_ES-1-FG484I') == info
    assert get_info_libero('MPF300TS_ES-FG484I-1') == info
    info = {
        'family': 'PolarFireSoC',
        'device': 'mpfs025t',
        'speed': 'STD',
        'package': 'fcvg484',
        'prange': 'EXT'
    }
    assert get_info_libero('MPFS025T-FCVG484E') == info
    info['prange'] = 'IND'
    assert get_info_libero('MPFS025T-FCVG484I') == info
    info['speed'] = '-1'
    info['prange'] = 'EXT'
    assert get_info_libero('MPFS025T-1FCVG484E') == info
    assert get_info_libero('MPFS025T-1-FCVG484E') == info
    assert get_info_libero('MPFS025T-FCVG484E-1') == info
    info['prange'] = 'IND'
    assert get_info_libero('MPFS025T-1FCVG484I') == info
    assert get_info_libero('MPFS025T-1-FCVG484I') == info
    assert get_info_libero('MPFS025T-FCVG484I-1') == info


def test_openflow():
    info = {'family': 'xc7', 'device': 'xc7k160t-3', 'package': 'fbg484'}
    assert get_info_openflow('xc7k160t-3-fbg484') == info
    info = {'family': 'ice40', 'device': 'hx1k', 'package': 'tq144'}
    assert get_info_openflow('hx1k-tq144') == info
    info = {'family': 'ecp5', 'device': '25k', 'package': 'CSFBGA285'}
    assert get_info_openflow('25k-CSFBGA285') == info
    info = {'family': 'ecp5', 'device': 'um5g-85k', 'package': 'CABGA381'}
    assert get_info_openflow('um5g-85k-CABGA381') == info



#!/usr/bin/env python3
#
# Copyright (C) 2020-2024 PyFPGA Project
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""
A CLI helper utility to deal with a vendor FPGA Project file.
"""

import argparse
import sys

from pathlib import Path
from pyfpga import __version__ as version
from pyfpga.factory import Factory, TOOLS
from pyfpga.project import STEPS

steps = list(STEPS.keys())[1:len(STEPS)]


def main():
    """Solves the main functionality of this helper."""

    # Parsing the command-line.

    parser = argparse.ArgumentParser(
        description=__doc__
    )

    parser.add_argument(
        '-v', '--version',
        action='version',
        version=f'v{version}'
    )

    parser.add_argument(
        '--last',
        metavar='STEP',
        choices=steps,
        default='bit',
        help=f'last step to perform [{steps[-1]}] ({"|".join(steps)})'
    )

    parser.add_argument(
        'prjfile',
        metavar='PRJFILE',
        help='a vendor Project File'
    )

    args = parser.parse_args()

    # Detecting a Project file

    tool_per_ext = {
        '.xise': 'ise',
        '.prjx': 'libero',
        '.qpf': 'quartus',
        '.xpr': 'vivado'
    }

    prjfile = Path(args.prjfile)

    if not prjfile.exists():
        sys.exit(f'ERROR: {prjfile} file not found.')

    directory = prjfile.parent
    base_name = prjfile.stem
    extension = prjfile.suffix

    tool = ''
    if extension in tool_per_ext:
        tool = tool_per_ext[extension]
        print(f'INFO: {tool} project file found.')
    else:
        sys.exit(f'ERROR: unknown project file extension ({extension})')

    # -------------------------------------------------------------------------
    # Solving with PyFPGA
    # -------------------------------------------------------------------------

    prj = Factory(tool, base_name, directory)

    try:
        prj.make('syn', args.last)
    except Exception as e:
        sys.exit('{} ({})'.format(type(e).__name__, e))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# Copyright (C) 2020-2024 PyFPGA Project
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""
A CLI helper utility to transform HDL design files into a bitstream.
"""

import argparse
import sys

from pathlib import Path
from pyfpga import __version__ as version
from pyfpga.factory import Factory, TOOLS
from pyfpga.project import STEPS

tools = list(TOOLS.keys())
steps = list(STEPS.keys())

EPILOGUE = f"""
Supported values of arguments with choices:
* TOOL = {'|'.join(tools)}
* STEP = {'|'.join(steps)}

Notes:
* PATH and FILE must be relative to the execution directory.
* The default PART name and how to specify it depends on the selected TOOL.
* More than one '--file', '--include' or '--param' arguments can be specified.
"""


def main():
    """Solves the main functionality of this helper."""

    # Parsing the command-line.

    parser = argparse.ArgumentParser(
        description=__doc__,
        epilog=EPILOGUE,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument(
        '-v', '--version',
        action='version',
        version=f'v{version}'
    )

    parser.add_argument(
        '-t', '--tool',
        metavar='TOOL',
        default='vivado',
        choices=tools,
        help='backend tool to be used [vivado]'
    )

    parser.add_argument(
        '-o', '--odir',
        metavar='PATH',
        default='results',
        help='where to generate files [results]'
    )

    parser.add_argument(
        '-p', '--part',
        metavar='PART',
        help='the target device'
    )

    parser.add_argument(
        '-f', '--file',
        metavar='FILE[,LIBRARY]',
        action='append',
        help='add a design file (optionally specifying a VHDL library)'
    )

    parser.add_argument(
        '-i', '--include',
        metavar='PATH',
        action='append',
        help='specify a Verilog Include directory'
    )

    parser.add_argument(
        '--define',
        metavar=('DEFINE', 'VALUE'),
        action='append',
        nargs=2,
        help='define and set the value of a Verilog Define'
    )

    parser.add_argument(
        '--param',
        metavar=('PARAMETER', 'VALUE'),
        action='append',
        nargs=2,
        help='set the value of a Generic/Parameter of the top-level'
    )

    parser.add_argument(
        '--project',
        metavar='PROJECT',
        help='optional PROJECT name'
    )

    parser.add_argument(
        '--last',
        metavar='STEP',
        choices=steps,
        default='bit',
        help=f'last step to perform [{steps[-1]}] ({"|".join(steps)})'
    )

    parser.add_argument(
        'toplevel',
        metavar='TOPLEVEL',
        help='the top-level name'
    )

    args = parser.parse_args()

    # -------------------------------------------------------------------------
    # Solving with PyFPGA
    # -------------------------------------------------------------------------

    project = args.project or args.tool

    prj = Factory(args.tool, project, odir=args.odir)

    if args.part is not None:
        prj.set_part(args.part)

    if args.include is not None:
        for include in args.include:
            prj.add_include(include)

    if args.file is not None:
        for file in args.file:
            aux = file.split(',')
            file = aux[0]
            lib = aux[1] if len(aux) > 1 else None
            ext = Path(file).suffix.lower()
            if ext == '.v':
                print(f'* Adding Verilog file: {file}')
                prj.add_vlog(file)
            elif ext == '.sv':
                print(f'* Adding System Verilog file: {file}')
                prj.add_slog(file)
            elif ext in ['.vhd', '.vhdl']:
                if lib:
                    print(f'* Adding VHDL file: {file} (library={lib})')
                else:
                    print(f'* Adding VHDL file: {file}')
                prj.add_vhdl(file, lib)
            else:
                print(f'* Adding Constraint file: {file}')
                prj.add_cons(file)

    if args.define is not None:
        for define in args.define:
            prj.add_define(define[0], define[1])

    if args.param is not None:
        for param in args.param:
            prj.add_param(param[0], param[1])

    prj.set_top(args.toplevel)

    try:
        prj.make(last=args.last)
    except Exception as e:
        sys.exit('{} ({})'.format(type(e).__name__, e))


if __name__ == "__main__":
    main()



#
# Copyright (C) 2019-2024 PyFPGA Project
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""
Implements support for Quartus.
"""

from pyfpga.project import Project


class Quartus(Project):
    """Class to support Quartus projects."""

    def _configure(self):
        tool = 'quartus'
        self.conf['tool'] = tool
        self.conf['make_cmd'] = f'quartus_sh --script {tool}.tcl'
        self.conf['make_ext'] = 'tcl'
        self.conf['prog_bit'] = ['sof', 'pof']
        self.conf['prog_cmd'] = f'bash {tool}-prog.tcl'
        self.conf['prog_ext'] = 'tcl'

    def _make_custom(self):
        if 'part' not in self.data:
            self.data['part'] = '10M50SCE144I7G'



#
# Copyright (C) 2024 PyFPGA Project
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""
Implements support for Diamond.
"""

import os
from pyfpga.project import Project


class Diamond(Project):
    """Class to support Diamond projects."""

    def _configure(self):
        tool = 'diamond'
        executable = 'pnmainc' if os.name == 'nt' else 'diamondc'
        self.conf['tool'] = tool
        self.conf['make_cmd'] = f'{executable} {tool}.tcl'
        self.conf['make_ext'] = 'tcl'
        self.conf['prog_bit'] = ['bit']
        self.conf['prog_cmd'] = f'sh {tool}-prog.sh'
        self.conf['prog_ext'] = 'sh'

    def _make_custom(self):
        if 'part' not in self.data:
            self.data['part'] = 'LFXP2-5E-5TN144C'



from setuptools import setup, find_packages

import pyfpga

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='PyFPGA',
    version=pyfpga.__version__,
    description='A Python package to use FPGA development tools programmatically',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Rodrigo A. Melo',
    author_email='rodrigomelo9@gmail.com',
    license='GPLv3',
    url='https://github.com/PyFPGA/pyfpga',
    package_data={'': ['templates/*.jinja', 'helpers/*']},
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'hdl2bit = pyfpga.helpers.hdl2bit:main',
            'prj2bit = pyfpga.helpers.prj2bit:main',
            'bitprog = pyfpga.helpers.bitprog:main'
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Utilities',
        'Topic :: Software Development :: Build Tools',
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)"
    ],
    install_requires=['jinja2']
)



#
# Copyright (C) 2019-2024 PyFPGA Project
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""
Base class that implements agnostic methods to deal with FPGA projects.
"""

import glob
import logging
import os
import subprocess

from pathlib import Path
from time import time
from jinja2 import Environment, FileSystemLoader


STEPS = {
    'cfg': 'Project Creation',
    'syn': 'Synthesis',
    'par': 'Place and Route',
    'bit': 'Bitstream generation'
}


class Project:
    """Base class to manage an FPGA project.

    :param project: project name (tool name when nothing specified)
    :type project: str, optional
    :param odir: output directory
    :type odir: str, optional
    """

    def __init__(self, project=None, odir='results'):
        """Class constructor."""
        self.conf = {}
        self.data = {}
        self._configure()
        self.data['project'] = project or self.conf['tool']
        self.odir = odir
        # logging config
        self.logger = logging.getLogger(self.__class__.__name__)
        if not self.logger.handlers:
            self.logger.setLevel(logging.INFO)
            handler = logging.StreamHandler()
            formatter = logging.Formatter(
                '%(asctime)s - %(levelname)s - %(message)s',
                datefmt='%Y-%m-%d %H:%M:%S'
            )
            handler.setFormatter(formatter)
            self.logger.addHandler(handler)

    def set_part(self, name):
        """Set the FPGA part name.

        :param name: FPGA part name
        :type name: str
        """
        self.logger.debug('Executing set_part: %s', name)
        self.data['part'] = name

    def add_include(self, path):
        """Add an Include path.

        Specify where to search for Included Verilog Files, IP repos, etc.

        :param path: path of a directory
        :type name: str
        :raises NotADirectoryError: if path is not a directory
        """
        self.logger.debug('Executing add_include: %s', path)
        path = Path(path).resolve()
        if not path.is_dir():
            raise NotADirectoryError(path)
        self.data.setdefault('includes', []).append(path.as_posix())

    def _add_file(self, pathname, hdl=None, lib=None):
        files = glob.glob(pathname, recursive=True)
        if len(files) == 0:
            raise FileNotFoundError(pathname)
        for file in files:
            path = Path(file).resolve().as_posix()
            attr = {}
            if hdl:
                attr['hdl'] = hdl
            if lib:
                attr['lib'] = lib
            if path in self.data.get('files', {}):
                del self.data['files'][path]
            self.data.setdefault('files', {})[path] = attr

    def add_slog(self, pathname):
        """Add System Verilog file/s.

        :param pathname: path to a SV file (glob compliant)
        :type pathname: str
        :raises FileNotFoundError: when pathname is not found
        """
        self.logger.debug('Executing add_slog: %s', pathname)
        self._add_file(pathname, 'slog')

    def add_vhdl(self, pathname, lib=None):
        """Add VHDL file/s.

        :param pathname: path to a SV file (glob compliant)
        :type pathname: str
        :param lib: VHDL library name
        :type lib: str, optional
        :raises FileNotFoundError: when pathname is not found
        """
        lib_str = 'default library' if lib is None else lib
        self.logger.debug('Executing add_vhdl: %s : %s', lib_str, pathname)
        self._add_file(pathname, 'vhdl', lib)

    def add_vlog(self, pathname):
        """Add Verilog file/s.

        :param pathname: path to a SV file (glob compliant)
        :type pathname: str
        :raises FileNotFoundError: when pathname is not found
        """
        self.logger.debug('Executing add_vlog: %s', pathname)
        self._add_file(pathname, 'vlog')

    def add_cons(self, path):
        """Add a constraint file.

        :param pathname: path of a file
        :type pathname: str
        :raises FileNotFoundError: if path is not found
        """
        self.logger.debug('Executing add_cons: %s', path)
        path = Path(path).resolve()
        if not path.is_file():
            raise FileNotFoundError(path)
        attr = {}
        self.data.setdefault('constraints', {})[path.as_posix()] = attr

    def add_param(self, name, value):
        """Add a Parameter/Generic Value.

        :param name: parameter/generic name
        :type name: str
        :param value: parameter/generic value
        :type name: str
        """
        self.logger.debug('Executing add_param: %s : %s', name, value)
        self.data.setdefault('params', {})[name] = value

    def add_define(self, name, value):
        """Add a Verilog Defile Value.

        :param name: define name
        :type name: str
        :param value: define value
        :type name: str
        """
        self.logger.debug('Executing add_define: %s : %s', name, value)
        self.data.setdefault('defines', {})[name] = value

    def add_fileset(self, pathname):
        """Add fileset file/s.

        :param pathname: path to a fileset file
        :type pathname: str
        :raises FileNotFoundError: when pathname is not found
        """
        self.logger.debug('Executing add_fileset: %s', pathname)
        if not os.path.exists(pathname):
            raise FileNotFoundError(pathname)
        raise NotImplementedError()

    def set_top(self, name):
        """Set the name of the top level component.

        :param name: top-level name
        :type name: str
        """
        self.logger.debug('Executing set_top: %s', name)
        self.data['top'] = name

    def add_hook(self, stage, hook):
        """Add a hook in the specific stage.

        A hook is a place that allows you to insert customized code.

        :param stage: where to insert the hook
        :type stage: str
        :param hook: a tool-specific command
        :type hook: str
        :raises ValueError: when stage is invalid
        """
        self.logger.debug('Executing add_hook: %s : %s', stage, hook)
        stages = [
            'precfg', 'postcfg', 'presyn', 'postsyn',
            'prepar', 'postpar', 'prebit', 'postbit'
        ]
        if stage not in stages:
            raise ValueError('Invalid stage.')
        self.data.setdefault('hooks', {}).setdefault(stage, []).append(hook)

    def set_debug(self):
        """Enables debug messages."""
        self.logger.setLevel(logging.DEBUG)

    def make(self, first='cfg', last='bit'):
        """Run the underlying tool.

        :param first: first step
        :type first: str, optional
        :param last: last step
        :type last: str, optional
        :raises ValueError: for missing or wrong values
        :raises RuntimeError: error running the needed underlying tool

        .. note:: valid steps are ``cfg``, ``syn``, ``par`` and ``bit``.
        """
        self.logger.debug('Executing make')
        if last not in STEPS:
            raise ValueError('Invalid last step.')
        if first not in STEPS:
            raise ValueError('Invalid first step.')
        keys = list(STEPS.keys())
        index = [keys.index(first), keys.index(last)]
        if index[0] > index[1]:
            raise ValueError('Invalid steps combination.')
        message = f'from {STEPS[first]} to {STEPS[last]}'
        if first == last:
            message = STEPS[first]
        self.logger.info('Running %s', message)
        self.data['steps'] = keys[index[0]:index[1]+1]
        self._make_custom()
        self._create_file(self.conf['tool'], self.conf['make_ext'])
        self._run(self.conf['make_cmd'], 'make.log')

    def prog(self, bitstream=None, position=1):
        """Program the FPGA

        :param bitstream: bitstream to be programmed
        :type bitstream: str, optional
        :param position: position of the device in the JTAG chain
        :type position: str, optional
        :raises ValueError: for missing or wrong values
        :raises FileNotFoundError: when bitstream is not found
        :raises RuntimeError: error running the needed underlying tool
        """
        self.logger.debug('Executing prog')
        if position not in range(1, 9):
            raise ValueError('Invalid position.')
        self.logger.info('Programming')
        if not bitstream:
            for ext in self.conf['prog_bit']:
                candidate = Path(self.odir) / f'{self.data["project"]}.{ext}'
                if candidate.exists():
                    bitstream = candidate.resolve()
                    break
        else:
            bitstream = Path(bitstream).resolve()
        if not bitstream or not bitstream.exists():
            raise FileNotFoundError(bitstream)
        self.data['bitstream'] = bitstream
        self._prog_custom()
        self._create_file(f'{self.conf["tool"]}-prog', self.conf['prog_ext'])
        self._run(self.conf['prog_cmd'], 'prog.log')

    def _configure(self):
        raise NotImplementedError('Tool-dependent')

    def _make_custom(self):
        pass

    def _prog_custom(self):
        pass

    def _create_file(self, basename, extension):
        tempdir = Path(__file__).parent.joinpath('templates')
        jinja_file_loader = FileSystemLoader(str(tempdir))
        jinja_env = Environment(loader=jinja_file_loader)
        jinja_template = jinja_env.get_template(f'{basename}.jinja')
        content = jinja_template.render(self.data)
        directory = Path(self.odir)
        directory.mkdir(parents=True, exist_ok=True)
        filename = f'{basename}.{extension}'
        with open(directory / filename, 'w', encoding='utf-8') as file:
            file.write(content)

    def _run(self, command, logname):
        num = 20
        error = 0
        old_dir = Path.cwd()
        new_dir = Path(self.odir)
        start = time()
        try:
            os.chdir(new_dir)
            with open(logname, 'w', encoding='utf-8') as file:
                subprocess.run(
                    command, shell=True, check=True, text=True,
                    stdout=file, stderr=subprocess.STDOUT
                )
        except subprocess.CalledProcessError:
            with open(logname, 'r', encoding='utf-8') as file:
                lines = file.readlines()
                last_lines = lines[-num:] if len(lines) >= num else lines
                for line in last_lines:
                    message = line.strip()
                    if len(message):
                        print(f'>> {message}')
            error = 1
        finally:
            os.chdir(old_dir)
            end = time()
            elapsed = end - start
            self.logger.info(
                'Elapsed time %dh %dm %.2fs',
                int(elapsed // 3600),
                int((elapsed % 3600) // 60),
                elapsed % 60
            )
            if error:
                raise RuntimeError('Problem with the underlying tool')



#!/usr/bin/env python3
#
# Copyright (C) 2020-2024 PyFPGA Project
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""
A CLI helper utility to transfer a bitstream to a supported device.
"""

import argparse
import sys

from pathlib import Path
from pyfpga import __version__ as version
from pyfpga.factory import Factory, TOOLS
from pyfpga.project import STEPS

tools = list(TOOLS.keys())
devs = ['fpga', 'spi', 'bpi']
positions = range(1, 10)
widths = [2**i for i in range(6)]

EPILOGUE = f"""
Supported values of arguments with choices:
* TOOL = {'|'.join(tools)}
* TYPE = {'|'.join(devs)}
* POSITION = {'|'.join(map(str, positions))}
* WIDTH = {'|'.join(map(str, widths))}
"""


def main():
    """Solves the main functionality of this helper."""

    # Parsing the command-line.

    parser = argparse.ArgumentParser(
        description=__doc__,
        epilog=EPILOGUE,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument(
        '-v', '--version',
        action='version',
        version=f'v{version}'
    )

    parser.add_argument(
        '-t', '--tool',
        metavar='TOOL',
        default='vivado',
        choices=tools,
        help='backend tool to be used [vivado]'
    )

    parser.add_argument(
        '-o', '--odir',
        metavar='PATH',
        default='results',
        help='where to generate files [results]'
    )

    parser.add_argument(
        '-d', '--device',
        metavar='TYPE',
        choices=devs,
        default=devs[0],
        help=f'the target device type [{devs[0]}]'
    )

    parser.add_argument(
        '-p', '--position',
        metavar='POSITION',
        choices=positions,
        type=int,
        default=1,
        help='the device position into the JTAG chain [1]'
    )

    parser.add_argument(
        '-m', '--memname',
        metavar='NAME',
        help='memory name for SPI or BPI devices [None]'
    )

    parser.add_argument(
        '-w', '--width',
        metavar='WIDTH',
        choices=widths,
        type=int,
        default=1,
        help='memory width for SPI or BPI devices [1]'
    )

    parser.add_argument(
        'bit',
        metavar='BITFILE',
        help='a bitstream file'
    )

    args = parser.parse_args()

    # -------------------------------------------------------------------------
    # Solving with PyFPGA
    # -------------------------------------------------------------------------

    prj = Factory(args.tool, odir=args.odir)

    try:
        if args.device == 'fpga':
            prj.prog(args.bit, args.position)
        if args.device == 'spi':
            prj.prog_spi(args.bit, args.position, args.width, args.memname)
        if args.device == 'bpi':
            prj.prog_bpi(args.bit, args.position, args.width, args.memname)
    except Exception as e:
        sys.exit('{} ({})'.format(type(e).__name__, e))


if __name__ == "__main__":
    main()
